'use strict';
const os = require('os')
const readline = require('readline')
const fs = require('fs')
const path = require('path')
const childProcess = require('child_process')
const util = require('util')
const crypto = require('crypto')
const getImageInfo = require('./getImageInfo.js')
const readDirR = require('./readDirR.js')
const md5 = (str) => crypto.createHash('md5').update(str).digest("hex")

const vips = path.join(__dirname, '..', `./soft/libs/vips/linux/vips`)
const vipsheader = path.join(__dirname, '..', `./soft/libs/vips/linux/vipsheader`)
const magick = path.join(__dirname, '..', `./soft/libs/imagemagick/linux/magick`)
const mozjpeg = path.join(__dirname, '..', `./soft/libs/mozjpeg/linux/jpegtran`)
const ect = path.join(__dirname, '..', `./soft/libs/ect/linux/ect`)

let pngMode = 'ect'

let dir = `./imgs/`
for (let arg of process.argv) {
    let match = arg.match(/-dir=['"]?(.*)['"]?/)
    if (match && match[1].trim()) {
        dir = match[1]
    }
}
console.log(`work dir: ${dir}`)

const asyncReadFile = util.promisify(fs.readFile)
const asyncStat = util.promisify(fs.stat)
const asyncExec = util.promisify(childProcess.exec)
const asyncCopyFile = util.promisify(fs.copyFile)
const asyncUnlink = util.promisify(fs.unlink)
const asyncRename = util.promisify(fs.rename)
const asyncChmod = util.promisify(fs.chmod)

const sizesEconom = {}

main()

async function checkPermission() {
    for (let item of [vips, vipsheader, magick, mozjpeg, ect]) {
        let err = await asyncChmod(item, 0o764)
        if (err) {
            console.error(err)
        }
    }
}

async function main() {
    await checkPermission()

    let to8Bits = false
    if (process.argv[2]) {
        pngMode = process.argv[2]
    }
    if (process.argv[4] === '8bit') {
        to8Bits = true
    }

    console.time('work time')
    const files = await readDirR(dir, { exts: ['.jpg', '.jpeg', '.png',] })

    try {
        await runQuery(files, optimizeFile, { to8Bits, pngMode })
    }
    catch (e) {
        console.error(e)
        askQuestion('some error')
        return
    }

    console.log('\n===========================================\n')
    console.log(calculateFinalSize(sizesEconom))
    console.timeEnd('work time')
    await askQuestion(`All ready`)
}

async function runQuery(files, fn, options) {
    let config = await readConfig()

    let accumulationIndex = 0
    let accumulationArr = []
    for (let i = 0; i < files.length; i++) {
        console.log(`${i + 1} / ${files.length}`)
        accumulationArr.push(fn(files[i], options))
        accumulationIndex += 1
        if (accumulationIndex >= config.cpu) {
            await Promise.all(accumulationArr)
            accumulationArr = []
            accumulationIndex = 0

            config = await readConfig()
        }
    }
    if (accumulationArr.length) {
        await Promise.all(accumulationArr)
    }
}

async function optimizeFile(filepathOrig, options = {}) {
    console.log(`run ${filepathOrig}`)
    options.to8Bits = options.to8Bits || false
    options.pngMode = options.pngMode || false

    const imageInfo = await getImageInfo(filepathOrig)

    if (!imageInfo) {
        return { status: 'error' }
    }

    let filepath = 'tmp_workfile' + md5(filepathOrig) + '.' + imageInfo.type

    try {
        await asyncCopyFile(filepathOrig, filepath)

        let stat = await asyncStat(filepath)
        sizesEconom[filepathOrig] = { before: stat.size, after: stat.size, }

        if (imageInfo.type === 'jpeg') {
            await optimizeJpg(filepath)
        }
        else if (imageInfo.type === 'png') {
            const reduceColor = process.argv[3] || false
            await optimizePng(filepath, options, reduceColor)
        }

        stat = await asyncStat(filepath)
        sizesEconom[filepathOrig].after = stat.size

        const percent = (100 - sizesEconom[filepathOrig].after * 100 / sizesEconom[filepathOrig].before).toFixed(2)
        const sizeBefore = (sizesEconom[filepathOrig].before / 1024 / 1024).toFixed(2)
        const sizeAfter = (sizesEconom[filepathOrig].after / 1024 / 1024).toFixed(2)
        const sizeSave = ((sizesEconom[filepathOrig].before - sizesEconom[filepathOrig].after) / 1024 / 1024).toFixed(2)
        console.log(`Size before: ${sizeBefore}mb, after: ${sizeAfter}mb. -${percent}% (${sizeSave}mb)`)

        if (sizesEconom[filepathOrig].before <= sizesEconom[filepathOrig].after) {
            await asyncUnlink(filepath)
        }
        else {
            await asyncRename(filepath, filepathOrig)
        }
    }
    catch (e) {
        console.error({ filename: filepathOrig, e })
    }
}

function calculateFinalSize(sizesEconom) {
    let beforeSize = 0
    let afterSize = 0
    for (let json of Object.values(sizesEconom)) {
        beforeSize += json.before
        afterSize += json.after
    }

    const sizeSave = ((beforeSize - afterSize) / 1024 / 1024).toFixed(2)

    let percent = (100 - afterSize * 100 / beforeSize).toFixed(2)
    return `Size before: ${(beforeSize / 1024 / 1024).toFixed(2)}mb, after: ${(afterSize / 1024 / 1024).toFixed(2)}mb. -${percent}% (${sizeSave}mb)`
}


function detectGray(pixels) {
    let falsePixels = 0
    let deltaSize = 30
    let isGrey = true
    for (let pixel of pixels) {
        if (!pixel.trim()) {
            continue
        }
        else if (pixel.match(/graya?\(\d+/)) {
            continue
        }
        else if (pixel.match(/srgb\(0,0,0\)/)) {
            continue
        }
        else {
            let match = pixel.match(/\(.*?(\d+).*?,.*?(\d+).*?,.*?(\d+).*?(,.*?\d+|).*?\)/)
            if (match) {
                let d1 = match[1]
                let d2 = match[2]
                let d3 = match[3]
                if (d1 === d2 && d2 === d3) {
                    continue
                }
                else if (Math.abs(d1 - d2) < deltaSize && Math.abs(d2 - d3) < deltaSize) {
                    continue
                }
            }
        }
        falsePixels += 1
        if (falsePixels > 5) {
            isGrey = false
            break
        }
    }
    return isGrey
}

async function optimizeJpg(filepath) {
    const cmd = `${mozjpeg} -optimize -progressive -outfile "${filepath}" "${filepath}"`
    await asyncExec(cmd)
}

async function convertTo8bits(filepath) {
    let imageInfo = await getImageInfo(filepath)
    if (imageInfo.interpretation !== 'srgb' && imageInfo.interpretation !== 'b-w') {
        let tempName = filepath + '____53wertwregtsdfg34gdfzzpng.png'
        let to8Bit = `"${vips}" colourspace "${filepath}" "${tempName}" srgb`
        await asyncExec(to8Bit)
        await asyncRename(tempName, filepath)
    }
}

async function convertToGrayIfGray(filepath, reduceColor) {
    let getColor = `${magick} convert "${filepath}" +dither -colors 256 -define histogram:unique-colors=true -format "%c" histogram:info:`
    let resultObj = await asyncExec(getColor)
    let result = resultObj.stdout
    result = result.toString().replace(/\r/, '').split('\n')
    let isGrey = detectGray(result)
    if (isGrey) {
        console.log('is grey:', isGrey)
        let tempName = filepath + '____53wertwregtsdfg34gdfzzpng.png'
        let toGray = `"${vips}" colourspace "${filepath}" "${tempName}" b-w`
        await asyncExec(toGray)
        await asyncRename(tempName, filepath)

        if (reduceColor && reduceColor > 0 && reduceColor < 256) {
            console.log(`reduce color to: ${reduceColor}`)
            let reduceColorCmd = `${magick} convert "${filepath}" +dither -colors ${reduceColor} "${filepath}"`
            await asyncExec(reduceColorCmd)
        }
    }
}

async function optimizePng(filepath, options, reduceColor) {
    if (options.to8Bits) {
        await convertTo8bits(filepath)
    }
    await convertToGrayIfGray(filepath, reduceColor)
    await asyncExec(`${ect} -2 -strip "${filepath}"`)
}

function askQuestion(query) {
    const rl = readline.createInterface({
        input: process.stdin,
        output: process.stdout,
    })

    return new Promise(resolve => rl.question(query, ans => {
        rl.close()
        resolve(ans)
    }))
}

async function readConfig() {
    const configFile = path.join(__dirname, '..', 'config.json')
    let data = await asyncReadFile(configFile)
    const config = JSON.parse(data)

    config.cpu = config.cpu || 'auto'
    if (config.cpu === 'auto') {
        config.cpu = os.cpus().length
    }
    config.cpu = parseInt(config.cpu)
    console.log(`cpu cnt: ${config.cpu}`)

    return config
}